import collections
import time

from coba.utilities import PackageChecker
from coba.simulations import Context, Action
from coba.learners.core import Learner, Key
from typing import Any, Dict, Sequence

class LinUCBLearner(Learner):
    """A learner using the LinUCB algorithm from "Contextual Bandits with Linear Payoff Functions" by Wei Chu et al."""

    @property
    def family(self) -> str:
        """The family of the learner.

        See the base class for more information
        """
        return f"linUCB"

    @property
    def params(self) -> Dict[str, Any]:
        """The parameters of the learner.

        See the base class for more information.
        """
        dict = {'alpha': self._alpha, 'interactions': self._interactions}
        return dict

    def __init__(self, *, alpha: float, interactions: Sequence[str] = ['a', 'ax'], timeit: bool = False) -> None:
        """Instantiate a linUCBLearner.
        Args:
            alpha: number of standard deviations
            interactions: the set of interactions the learner will use. x refers to context and a refers to actions, 
                e.g. xaa would mean interactions between context and actions and actions. 
        """
        PackageChecker.numpy("linUCBLearner.__init__")

        self._A            = None
        self._b            = None
        self._alpha        = alpha
        self._interactions = interactions
        self._terms        = []
        self._times        = [0.,0.]
        self._i            = 0
        self._timeit       = timeit

        for term in self._interactions:
            term = term.lower()
            x_num = term.count('x')
            a_num = term.count('a')
            
            if x_num + a_num != len(term):
                raise Exception("Letters other than x and a were passed for parameter interactions. Please remove other letters/characters.")
            
            self._terms.append((x_num, a_num))

    def predict(self, key: Key, context: Context, actions: Sequence[Action]) -> Sequence[float]:
        """Determine a PMF with which to select the given actions.

        Args:
            key: The key identifying the interaction we are choosing for.
            context: The context we're currently in. See the base class for more information.
            actions: The actions to choose from. See the base class for more information.

        Returns:
            The probability of taking each action. See the base class for more information.
        """
        import numpy as np #type: ignore

        self._i += 1

        self._d   = len(actions[0]) if isinstance(actions[0], collections.Sequence) else 1
        is_sparse = isinstance(actions[0], dict) or isinstance(context, dict)

        if is_sparse:
            raise Exception("Sparse data cannot be handled by this algorithm.")

        features: np.ndarray = self._featurize(context, actions)

        if(self._A is None):
            self._A = np.identity(features.shape[0])
            self._b = np.zeros((features.shape[0], 1))
        
        start_predict = time.time()

        A_inv = np.linalg.inv(self._A)

        self._times[0] += time.time() - start_predict

        theta = np.dot(A_inv, self._b)
        
        term_one = np.zeros([len(actions),1])
        term_two = np.zeros([len(actions),1])

        for i in range(len(actions)):
            term_one[i] = theta.T @ features[:,i]
            term_two[i] = self._alpha * np.sqrt(features[:,i].T @ A_inv @ features[:,i])

        action_values = term_one + term_two

        if (self._i-1) % 100 == 0 and self._timeit:
            print(self._times[0]/(self._i+1))
            print(self._times[1]/(self._i+1))

        max_indexes = np.where(action_values == np.amax(action_values))[0]
        return [1/len(max_indexes) if ind in max_indexes else 0 for ind in range(len(actions))]

    def learn(self, key: Key, context: Context, action: Action, reward: float, probability: float) -> None:
        """Learn from the given interaction.

        Args:
            key: The key identifying the interaction this observed reward came from.
            context: The context we're learning about. See the base class for more information.
            action: The action that was selected in the context. See the base class for more information.
            reward: The reward that was gained from the action. See the base class for more information.
            probability: The probability that the given action was taken.
        """
        import numpy as np #type: ignore

        learn_start = time.time()
        
        features: np.ndarray = self._featurize(context, [action])

        self._A = self._A + features@features.T
        self._b = self._b + features*reward 

        self._times[1] += time.time() - learn_start

    def _featurize(self, context, actions):
        import numpy as np #type: ignore

        features = np.array([[]]*len(actions))

        context_array = np.array(context) if context is not None else np.array([[1]])
        actions_array = np.array(actions)

        for term in self._terms:
            temp_array_1 = [[1.0]]
            temp_array_2 = np.ones((len(actions), 1 ), dtype=float)

            for _ in range(term[0]):
                temp_array_1 = np.outer(temp_array_1, context_array).reshape(-1)
            
            for _ in range(term[1]):
                temp_array_2 = [np.outer(temp_array_2[i], actions_array[i]).reshape(-1) for i in range(len(actions_array))]
            
            temp_array = np.apply_along_axis(lambda x: np.outer(x, temp_array_1).reshape(-1), 1, temp_array_2)

            features = np.hstack([features, temp_array])

        return features.T


"""The public API for the config module.

This module contains coba configuration functionality.
"""

from coba.config.core       import CobaConfig
from coba.config.cachers    import NoneCacher, MemoryCacher, DiskCacher, Cacher
from coba.config.loggers    import NoneLogger, BasicLogger, IndentLogger, Logger
from coba.config.exceptions import CobaException, CobaFatal

__all__ =[
    'CobaConfig',
    'CobaException',
    'CobaFatal',
    'NoneCacher',
    'MemoryCacher',
    'DiskCacher',
    'Cacher',
    'NoneLogger',
    'BasicLogger',
    'IndentLogger',
    'Logger'
]


from coba.simulations.core import Interaction
from copy import deepcopy
from itertools import groupby, product, count
from collections import defaultdict
from typing import Iterable, Sequence, Any, Optional, Dict, Hashable

from coba.random import CobaRandom
from coba.learners import Learner
from coba.config import CobaConfig
from coba.pipes import Pipe, Filter, Source, IdentityFilter
from coba.simulations import Context, Action, Key, Simulation, SimulationFilter

from coba.benchmarks.transactions import Transaction
from coba.benchmarks.results import Result

class BenchmarkTask:

    class BenchmarkTaskLearner(Learner):

        @property
        def family(self) -> str:
            try:
                return self._learner.family
            except AttributeError:
                return self._learner.__class__.__name__

        @property
        def params(self) -> Dict[str, Any]:
            try:
                return self._learner.params
            except AttributeError:
                return {}

        def __init__(self, learner: Learner, seed: Optional[int]) -> None:
            self._learner = learner
            self._random  = CobaRandom(seed)

        def predict(self, key: Key, context: Context, actions: Sequence[Action]) -> Sequence[float]:
            return self._learner.predict(key, context, actions) #type: ignore

        def learn(self, key: Key, context: Context, action: Action, reward: float, probability: float) -> Optional[Dict[str,Any]]:
            return self._learner.learn(key, context, action, reward, probability) #type: ignore

    class BenchmarkTaskSimulation(Simulation):

        def __init__(self, pipe: Simulation) -> None:
            self._pipe = pipe

        @property
        def source(self) -> Simulation:
            return self._pipe._source if isinstance(self._pipe, (Pipe.SourceFilters)) else self._pipe

        @property
        def filter(self) -> SimulationFilter:
            return self._pipe._filter if isinstance(self._pipe, Pipe.SourceFilters) else IdentityFilter()

        def read(self) -> Iterable[Interaction]:
            return self._pipe.read()

        def __repr__(self) -> str:
            return self._pipe.__repr__()

    def __init__(self, src_id:int, sim_id: int, lrn_id: int, simulation: Source[Simulation], learner: Learner, seed: int = None) -> None:
        self.src_id     = src_id
        self.sim_id     = sim_id
        self.lrn_id     = lrn_id
        
        self.seed       = seed
        self.simulation = BenchmarkTask.BenchmarkTaskSimulation(simulation)
        self.learner    = BenchmarkTask.BenchmarkTaskLearner(learner, seed)

class Tasks(Source[Iterable[BenchmarkTask]]):

    def __init__(self, simulations: Sequence[Source[Simulation]], learners: Sequence[Learner], seed: int = None) -> None:
        self._simulations = simulations
        self._learners    = learners
        self._seed        = seed

    def read(self) -> Iterable[BenchmarkTask]:

        #we rely on sim_id to make sure we don't do duplicate work. So long as self._simulations
        #is always in the exact same order we should be fine. In the future we may want to consider.
        #adding a better check for simulations other than assigning an index based on their order.

        source_ids: Dict[Hashable, int]  = defaultdict(lambda x=count(): next(x)) # type: ignore

        for (sim_id,sim), (lrn_id,lrn) in product(enumerate(self._simulations), enumerate(self._learners)):
            sim_source = sim._source if isinstance(sim, (Pipe.SourceFilters)) else sim
            yield BenchmarkTask(source_ids[sim_source], sim_id, lrn_id, sim, lrn, self._seed)                

class Unfinished(Filter[Iterable[BenchmarkTask], Iterable[BenchmarkTask]]):
    def __init__(self, restored: Result) -> None:
        self._restored = restored

    def filter(self, tasks: Iterable[BenchmarkTask]) -> Iterable[BenchmarkTask]:

        def is_not_complete(sim_id: int, learn_id: int):
            return (sim_id,learn_id) not in self._restored._interactions

        for task in tasks:
            if is_not_complete(task.sim_id, task.lrn_id):
                yield task

class ChunkBySource(Filter[Iterable[BenchmarkTask], Iterable[Iterable[BenchmarkTask]]]):

    def filter(self, tasks: Iterable[BenchmarkTask]) -> Iterable[Iterable[BenchmarkTask]]:

        srt_key = lambda t: t.src_id
        grp_key = lambda t: t.src_id

        tasks = list(tasks)

        for _, group in groupby(sorted(tasks, key=srt_key), key=grp_key):
            yield list(group)

class ChunkByTask(Filter[Iterable[BenchmarkTask], Iterable[Iterable[BenchmarkTask]]]):

    def filter(self, tasks: Iterable[BenchmarkTask]) -> Iterable[Iterable[BenchmarkTask]]:

        for task in tasks:
            yield [ task ]

class ChunkByNone(Filter[Iterable[BenchmarkTask], Iterable[Iterable[BenchmarkTask]]]):

    def filter(self, tasks: Iterable[BenchmarkTask]) -> Iterable[Iterable[BenchmarkTask]]:
        yield list(tasks)

class Transactions(Filter[Iterable[Iterable[BenchmarkTask]], Iterable[Any]]):

    def filter(self, chunks: Iterable[Iterable[BenchmarkTask]]) -> Iterable[Any]:

        for chunk in chunks:
            for transaction in self._process_chunk(chunk):
                yield transaction

    def _process_chunk(self, task_group: Iterable[BenchmarkTask]) -> Iterable[Any]:

        source_by_id = { t.src_id: t.simulation.source for t in task_group }
        filter_by_id = { t.sim_id: t.simulation.filter for t in task_group }

        srt_src = lambda t: t.src_id
        grp_src = lambda t: t.src_id
        srt_sim = lambda t: t.sim_id
        grp_sim = lambda t: t.sim_id

        with CobaConfig.Logger.log(f"Processing chunk..."):

            for src_id, tasks_by_src in groupby(sorted(task_group, key=srt_src), key=grp_src):

                try:

                    with CobaConfig.Logger.time(f"Creating source {src_id} from {source_by_id[src_id]}..."):
                        #Rhis is not ideal. I'm not sure how it should be improved and leaving this for now.
                        loaded_source = list(source_by_id[src_id].read())

                    for sim_id, tasks_by_src_sim in groupby(sorted(tasks_by_src, key=srt_sim), key=grp_sim):

                        tasks_by_src_sim_list = list(tasks_by_src_sim)
                        learner_ids           = [t.lrn_id  for t in tasks_by_src_sim_list]
                        learners              = [t.learner for t in tasks_by_src_sim_list]
                        seeds                 = [t.seed    for t in tasks_by_src_sim_list]

                        learner_ids.reverse()
                        learners.reverse() 

                        with CobaConfig.Logger.time(f"Creating simulation {sim_id} from source {src_id}..."):
                            interactions = filter_by_id[sim_id].filter(loaded_source)

                        if not interactions:
                            CobaConfig.Logger.log(f"Simulation {sim_id} has nothing to evaluate (likely due to `take` being larger than the simulation).")
                            continue

                        for index in sorted(range(len(learners)), reverse=True):

                            lrn_id  = learner_ids[index]
                            learner = deepcopy(learners[index])
                            random  = CobaRandom(seeds[index])

                            try:
                                with CobaConfig.Logger.time(f"Evaluating learner {lrn_id} on Simulation {sim_id}..."):

                                    row_data = defaultdict(list)

                                    for i, interaction in enumerate(interactions):
                                        probs  = learner.predict(i, interaction.context, interaction.actions)
                                        
                                        assert abs(sum(probs) - 1) < .0001, "The learner returned invalid proabilities for action choices."
                                        
                                        action = random.choice(interaction.actions, probs)
                                        reward = interaction.feedbacks[interaction.actions.index(action)]
                                        prob   = probs[interaction.actions.index(action)]
                                        
                                        info = learner.learn(i, interaction.context, action, reward, prob) or {}
                                                                                
                                        for key,value in info.items() | {('reward',reward)}: 
                                            row_data[key].append(value)

                                    yield Transaction.interactions(sim_id, lrn_id, _packed=row_data)

                            except Exception as e:
                                CobaConfig.Logger.log_exception(e)

                            finally:
                                del learner_ids[index]
                                del learners[index]

                except Exception as e:
                    CobaConfig.Logger.log_exception(e)